(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_arith.mli,v 1.28 2001/06/19 13:05:39 barnier Exp $ *)

(* Module [Arith]: Arithmetic Expressions over Variables of Type [Var.Fd.t] *)

(*** Overview *)
(*q This module provides functions and operators to build arithmetic
   expressions and state (in/dis)equation constraints over them.
*)  

(*** Basic *)
type t
(* Type of arithmetic expressions over variables of type [Var.Fd.t] and
   integers. *)
val fprint : out_channel -> t -> unit
(* [fprint chan e] prints expression [e] on channel [chan]. *)
val eval : t -> int
(* [eval e] returns the integer numerical value of a fully instantiated
   expression [e]. Raises [Invalid_argument] if [e] is not instantiated. *)
val min_of_expr : t -> int
val max_of_expr : t -> int
(* [min_of_expr e] (resp. [max_of_expr e]) returns the minimal (resp. maximal)
   possible value of expression [e]. *)

(*** Conversion *)
val i2e : int -> t
(* [i2e n] returns an expression which evaluates to [n]. *)
val fd2e : Fcl_var.Fd.t -> t
(* [fd2e v] returns an expression which evaluates to [n] when [v] is
   instantiated and [Var.Fd.value v] evaluates to [Val n]. *)
val e2fd : t -> Fcl_var.Fd.t
(* [e2fd e] creates and returns a new variable [v] and posts the constraint
   [fd2e v =~ e]. *)

(*** Construction of Arithmetic Expressions *)
(*q
  The arithmetic operators check whether any integer overflow (i.e. the
   result of an arithmetic operation over integers is less than [min_int]
   or greater than [max_int]) occurs during constraints internal computations
   and raise an assert failure only when using the byte code library
  [facile.cma]. *)
val (+~) : t -> t -> t
val (-~) : t -> t -> t
val ( *~) : t -> t -> t
(* Addition, substraction, multiplication over expressions. *)
val ( **~) : t -> int -> t
(* Exponentiation of an expression to an integer value. *)
val (/~) : t -> t -> t
val (%~) : t -> t -> t
(* Division and modulo over expressions. The exception [Division_by_zero]
   is raised whenever the second argument evaluates to 0. *)
val abs : t -> t
(* Absolute value over expressions. *)
val sum : t array -> t
val sum_fd : Fcl_var.Fd.t array -> t
(* [sum exps] (resp. [sum_fd vars]) returns the sum of all the elements of an
   array of expressions (resp. variables). Returns [Fd.int 0]
   if the array is empty. *)
val scalprod : int array -> t array -> t
val scalprod_fd : int array -> Fcl_var.Fd.t array -> t
(* [scalprod coeffs exps] (resp. [scalprod_fd coeffs vars]) returns the
   scalar product of an array of integers and an array of expressions
   (resp. variables). Raises [Invalid_argument] if the arrays have not the
   same length. *)
val prod : t array -> t
val prod_fd : Fcl_var.Fd.t array -> t
(* [prod exps] (resp. [prod_fd vars]) returns the product of all the
   elements of an array of expressions (resp. variables). *)

(*** Arithmetic Constraints over Expressions *)
val (<~)  : t -> t -> Fcl_cstr.t
val (<=~) : t -> t -> Fcl_cstr.t
val (=~)  : t -> t -> Fcl_cstr.t
val (>=~) : t -> t -> Fcl_cstr.t
val (>~)  : t -> t -> Fcl_cstr.t
val (<>~)  : t -> t -> Fcl_cstr.t
(* Constraints strictly less, less or equal, equal, greater or equal,
   strictly greater and different over expressions. *)

(*** Reification *)
(*q 
   Reification of the arithmetic constraint which is replaced by an expression
   equal to a boolean variable instantiated to [1] if the constraint is
   satisfied and to [0] if it is violated. These operators are shortcuts
   to lighten the writing of reified expressions:*)

(* [e1 op~~ e2] is equivalent to [fd2e (Reify.boolean (e1 op~ e2))]. *)
val (<~~)  : t -> t -> t
val (<=~~) : t -> t -> t
val (=~~)  : t -> t -> t
val (>=~~) : t -> t -> t
val (>~~)  : t -> t -> t
val (<>~~)  : t -> t -> t
(* Reified strictly less, less or equal, equal, greater or equal,
   strictly greater and different. *)
