(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_domain.mli,v 1.33 2001/06/20 16:09:40 barnier Exp $ *)

(* Module [Domain]: Domain Operations *)

(* This module provides all necessary functions (and more) to create
   and handle domains, which are needed to create variables and perform
   propagation (i.e. domain reduction or filtering). *)
   

type t
(* Type of finite domains of integers (functional: no in place
   modifications, domains can be shared). Standard equality and
   comparison can be used on type domain. *)

(*** Building New Domains *)
val empty : t
(* The empty domain. *)
val create : int list -> t
(* [create l] builds a new domain containing the values of [l]. Removes
   duplicates and sorts values. Returns [empty] if [l] is
   empty. *)
val unsafe_create : int list -> t
(* [unsafe_create l] builds a new domain containing the values of [l]. [l] must
be sorted and must not contain duplicate values, otherwise behaviour is
unspecified. Returns [empty] if [l] is empty. It is a more efficient variant
of [create]. *)
val interval : int -> int -> t
  (* [interval inf sup] returns the domain of all integers in the closed
     interval [[inf..sup]]. Raise [Invalid_argument] if [inf > sup]. *)
val int : t
  (* The largest representable domain, handy to create variables for which
     bounds cannot be previously known. It is actually much smaller
     than [interval min_int max_int] (which really is the biggest one) to
     help to avoid overflows while computing bounds of expressions
     involving such variables. *)
val boolean : t
  (* The domain containing [0] and [1]. *)

(*** Access *)
val is_empty : t -> bool
(* [is_empty d] tests if the domain [d] is empty. *)
val size : t -> int
(* [size d] returns the number of integers in [d]. *)
val min : t -> int
val max : t -> int
(* [min d] (resp. [max d]) returns the lower (resp. upper) bound of [d].
   If [d] is empty, the behaviour is unspecified (incorrect return value
   or exception raised). *)
val min_max : t -> int * int
(* [min_max d] returns both the lower and upper bound of [d]. If [d] is empty,
   the behaviour is unspecified (incorrect return value or exception
   raised). *)
val iter : (int -> unit) -> t -> unit
  (* [iter f d] successively applies function [f] to all element of [d] in
  increasing order. *)
val interval_iter : (int -> int -> unit) -> t -> unit
  (* [interval_iter f d] successively applies function [f] to the bounds
     of all the disjoint intervals of [d] in increasing order. E.g. a
     suitable function to print a domain can be
     [fun mini maxi -> Printf.printf "%d..%d " mini maxi]. *)
val member : int -> t -> bool
(* [member n d] tests if [n] belongs to [d]. *)
val values : t -> int list
  (* [value d] returns the list of values of the domain [d] *)
val fprint : out_channel -> t -> unit
(* [fprint chan d] prints [d] on channel [chan]. *)
val sprint : t -> string
(* [sprint d] returns a string representation of [d]. *)
val included : t -> t -> bool
  (* [included d1 d2] tests if domain [d1] is included in domain [d2]. *)
val smallest_geq : t -> int -> int
val greatest_leq : t -> int -> int
  (* [smallest_geq dom val] (resp. [greatest_leq dom val]) returns the
     smallest (resp. greatest) value in [dom] greater (resp. smaller) than
     or equal to [val]. Raises [Not_found] if [max dom < val] (resp.
     [min dom > val]). *)
val choose : (int -> int -> bool) -> t -> int
  (* [choose ord d] returns the mininum value of [d] for order [ord].
     Raises [Not_found] if [d] is empty. *)

(*** Operations *)
val remove : int -> t -> t
(* [remove n d] returns [d]-{[n]}. Returns [d] if [n] is not in [d]. *)
val add : int -> t -> t
(* [add n d] returns [d]+{[n]}. *)
val remove_up : int -> t -> t
val remove_low : int -> t -> t
  (* [remove_up n d] (resp. [remove_low n d]) returns [d]-[[n+1..max_int]]
     (resp. [d]-[[min_int..n-1]]), i.e. removes values
     stricly greater (resp. less) than [n]. *)
val remove_closed_inter : int -> int -> t -> t
  (* [remove_closed_inter inf sup d] returns [d]-[[inf..sup]], i.e. removes
     values greater than or equal to [inf] and less or equal to [sup] in [d].
     Returns [d] if [inf > sup]. *)
val intersection : t -> t -> t
val union : t -> t -> t
(* Intersection (resp. union) over domains. *)
val difference : t -> t -> t
    (* [difference big small] returns [big]-[small]. [small] must be included
       in [big], otherwise the behaviour is unspecified (incorrect return
       value or exception raised). *)
val remove_min : t -> t
val remove_max : t -> t
(* [remove_min d] (resp. [remove_max d]) returns [d] without its lower
   (resp. upper) bound. Raises [Invalid_argument] if [d] is empty. *)
val minus : t -> t
  (* [minus d] returns the domain of opposite values of [d]. *)
val plus : t -> int -> t
  (* [plus d n] translates a domain by [n]. *)

