(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                 A Functional Constraint Library                     *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, CENA               *)
(*                                                                     *)
(* Copyright 2001 CENA. All rights reserved. This file is distributed  *)
(* under the terms of the GNU Lesser General Public License.           *)
(***********************************************************************)
(* $Id: fcl_goals.ml,v 1.37 2001/06/07 15:44:30 barnier Exp $ *)

open Fcl_var

class virtual t = Fcl_stak.gl

let save = Fcl_stak.save

exception Success

let goal_stack = ref [];;

let reset () = Fcl_stak.reset (); Fcl_cstr.reset_queue ();;

let while_true = fun _ -> ()

let and_loop () =
  while true do
    match !goal_stack with
      [] -> raise Success
    | goal::goals ->
        goal_stack := goals;
        begin
	  match goal#call with
	    None -> ()
	  | Some g -> goal_stack := g :: !goal_stack
	end
  done

let on_choice_point = Fcl_cstr.new_event ()

let solve ?(control = while_true) goal =
  Fcl_misc.protect "Goals.solve"
    (fun () ->
      Fcl_stak.reset ();
      let backtracks = ref 0 in
      goal_stack := [goal];
      try
	while true do (* OR loop *)
    	  try
            control !backtracks;
	    Fcl_cstr.schedule on_choice_point;
	    Fcl_cstr.wake_all ();
	    Fcl_cstr.reset_queue ();
	    and_loop ()
    	  with
	    Fcl_stak.Fail s ->
              Fcl_debug.call 'g' (fun f ->  Printf.fprintf f "fail %s\n" s);
              goal_stack := Fcl_stak.backtrack (); (* May raise Empty_stack *)
              incr backtracks
	done;
	failwith "end of while true"
      with
    	Fcl_stak.Empty_stack -> false
      | Success -> control !backtracks; true);;


class create_rec name f = object (self)
  inherit t
  method name = name
  method call = Some (f self)
end
let create_rec ?(name = "create_rec") f = new create_rec name f

class create name f a = object
  inherit t
  method name = name
  method call = Some (f a)
end
let create ?(name = "create") f a = new create name f a

class atomic name f = object
  inherit t
  method name = name
  method call = f (); None
end
let atomic ?(name = "atomic") f = new atomic name f

class success = object
  inherit t
  method name = "success"
  method call = None
end
let success = new success

class fail = object
  inherit t
  method name = "fail"
  method call = Fcl_stak.fail "fail"; None
end
let fail = new fail

let (&&~) g1 g2 =
  create ~name:"&&~" (fun () -> goal_stack := g2:: !goal_stack; g1) ()
let (||~) g1 g2 =
  create ~name:"||~" (fun () -> ignore(Fcl_stak.save (g2:: !goal_stack)); g1) ()

let indomain var =
  create_rec ~name:"indomain"
    (fun self ->
      match Fd.value var with
      Val _ -> success
    | Unk var_ -> 
    	let mini = Fcl_domain.min (Attr.dom var_) in
	  atomic (fun () -> Fd.subst var mini) ||~
	  atomic
	    (fun () -> Fd.refine var (Fcl_domain.remove mini (Attr.dom var_)))
	    &&~ self);;
 (* can t use remove_min because min may have been changed in the
    choice point (Opti.minimize Continue does that) *)

let dichotomic var =
  create_rec ~name:"dichotomic"
    (fun self ->
      match Fd.value var with
      	Val _ -> success
      | Unk var_ -> 
      	  let d = Attr.dom var_ in
      	  let mini = Fcl_domain.min d and maxi = Fcl_domain.max d in
      	  let middle = (maxi + mini) / 2 in
	  (atomic (fun () -> Fd.refine var (Fcl_domain.remove_up middle d)) ||~
	  atomic (fun () -> Fd.refine var (Fcl_domain.remove_low (middle+1) d))) &&~
	  self)

let instantiate choose var =
  create_rec ~name:"instantiate"
    (fun self ->
      match Fd.value var with
      Val _ -> success
    | Unk var_ -> 
    	let x = choose (Attr.dom var_) in
	atomic (fun () -> Fd.subst var x) ||~
	atomic (fun () -> Fd.refine var (Fcl_domain.remove x (Attr.dom var_))) &&~ self);;


let once goal =
  let l = ref (Fcl_stak.level ()) in
  atomic (fun () -> l := (Fcl_stak.level ())) &&~
  goal &&~
  atomic (fun () -> Fcl_stak.cut !l)

let forto min max g =
  let rec la i =
    create
      (fun j ->
	 if j <= max then g j &&~ la (j+1) else success)
      i in
  la min

let fordownto max min g =
  forto min max (fun i -> g (max - i + min));;

module Array = struct
  let fold_hi select lab_one a init =
    create_rec
      (fun self ->
	try let i = select a in lab_one i a.(i) self with Not_found -> init)

  let foldi lab_one a init =
    let size = Array.length a in
    let rec la i =
      create
	(fun j -> if j < size then lab_one j a.(j) (la (j+1)) else init)
	i in
    la 0

  let foralli ?select f a =
    match select with
      None -> foldi (fun i x r -> f i x &&~ r) a success
    | Some s -> fold_hi s (fun i x r -> f i x &&~ r) a success

  let forall ?select f a = foralli ?select (fun _ -> f) a

  let existsi ?select f a =
    match select with
      None -> foldi (fun i x r -> f i x ||~ r) a fail
    | Some s -> fold_hi s (fun i x r -> f i x ||~ r) a fail

  let exists ?select f a = existsi ?select (fun _ -> f) a

  let choose_index order tab =
    let n = Array.length tab in

    (* Recherche de la premiere variable libre *)
    let rec first_unbound i = 
      if i < n then
      	match Fd.value tab.(i) with
	  Unk attr -> i, attr
	| Val _ -> first_unbound (i+1)
      else
	raise Not_found in
    
    let b, attr = first_unbound 0 in
    let best = ref b and attr_best = ref attr in
    (* Recherche de la meilleure variable pour le critere *)
    for i = b+1 to n - 1 do
      match Fd.value tab.(i) with
	Unk tabi ->
	  if order tabi !attr_best then begin
	    best := i; attr_best := tabi
	  end
      |	Val _ -> ()
    done;
    !best

  let labeling (a : Fcl_var.Fd.t array) = forall indomain a

  exception Return of int
  let not_instantiated_fd fds =
    try
      Array.iteri
      ~f:(fun i fdsi -> if Fd.is_var fdsi then raise (Return i))
	fds;
      raise Not_found
    with
      Return i -> i
end

module List = struct
  let rec fold fgoal l init =
    create
      (function
	   [] -> init
	 | x::xs -> fgoal x (fold fgoal xs init))
      l

  let rec fold_h select fgoal l init =
    create
      (function
	  [] -> init
       	| _ ->
            let x,xs = select l in
            fgoal x (fold_h select fgoal xs init))
      l

  let forall ?select f l =
    match select with
      None -> fold (fun x r -> f x &&~ r) l success
    | Some s -> fold_h s (fun x r -> f x &&~ r) l success

  let exists ?select f l =
    match select with
      None -> fold (fun x r -> f x ||~ r) l fail
    | Some s -> fold_h s (fun x r -> f x ||~ r) l fail

  let member v l = exists (fun x -> atomic (fun () -> Fd.unify v x)) l

  let labeling = forall indomain
end

let save = Fcl_stak.save

let unify v x = atomic (fun () -> Fd.unify v x)

open Fcl_arith

let level g =
  create (fun () -> let l = Fcl_stak.level () in g l) ()

let sigma ?(domain = Fcl_domain.int) g =
  create (fun () -> g (Fd.create domain)) ()

let once g =
  level (fun l -> g &&~ atomic (fun () -> Fcl_stak.cut l))

let minimize_restart goal (cost : Fd.t) step compute_solution =
  let best_cost = ref 0
  and once_more = ref true in
  atomic
    (fun () ->
      best_cost := Fd.max cost + step;
      once_more := true)
    &&~
  create_rec
    (fun self ->
      if !once_more then (* First iteration or last one succeeded *)
      	let loop_level = ref (Fcl_stak.level ()) in
	once_more := false;
      	(atomic (* Stores the choice-point and constrains the cost *)
	   (fun () ->
	     loop_level := Fcl_stak.level ();
	     Fcl_cstr.post (fd2e cost <=~ i2e (!best_cost - step)))
	   &&~
	 goal (* Solves the goal *)
 	   &&~
	 atomic
	   (fun () -> (* One solution found *)
	     once_more := true;
  (* Dire dans la doc que goal doit obligatoirement instancier le cout *)
	     let m = Fd.int_value cost in
	     compute_solution m;
	     best_cost := m;
	     Fcl_stak.cut !loop_level; raise (Fcl_stak.Fail "Goals.minimize")))
        ||~
	  self
      else (* Last try failed *)
	fail)


let minimize_continue goal (cost : Fd.t) step compute_solution =
  let rec bt_until c l =
    (* Backtrack until lower bound better than current cost, staying above [l] *)
    let gs = Fcl_stak.backtrack () in
    if Fd.min cost < c || Fcl_stak.older (Fcl_stak.level ()) l then begin
      ignore (Fcl_stak.save gs)
    end else
      bt_until c l in

  let best_cost = ref 0 in

  let restore_max =
    let update _ =
(***       Printf.printf "cost=%a best_cost=%d\n" Fd.fprint cost !best_cost; flush stdout; ***)
      match Fd.value cost with
	Val v -> if v > !best_cost then Fcl_stak.fail "restore_max" else true
      | Unk attr ->
	  Fd.refine cost (Fcl_domain.remove_up !best_cost (Attr.dom attr));
	  false
    and delay x = Fcl_cstr.delay [on_choice_point] x in
    Fcl_cstr.create ~name:"restore_cost" update delay in

  let found_one l =
    atomic ~name:"found_one"
      (fun () ->
        let c = Fd.int_value cost in
      	compute_solution c;
	best_cost := c - step;
	bt_until c l;
      	Fcl_stak.fail "Goals.minimize_more") in

  let init =
    atomic
      (fun () ->
 	best_cost := Fd.max cost + 1;
	Fcl_cstr.post restore_max) in

  level (fun l -> init &&~ goal &&~ found_one l);;

type bb_mode = Restart | Continue

let minimize ?(step=1) ?(mode = Continue) g c cs =
  if step <= 0 then invalid_arg "Goals.minimize: step must be non negative";
  match mode with
    Restart -> minimize_restart g c step cs
  | Continue -> minimize_continue g c step cs
